﻿using System;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class SocketUdpServer
    {
        private readonly UdpClient _udpClient;

        public bool IsIPv6 { get; private set; }

        public event Action<RemoteInfo> OnReceive;
        public event Action<RemoteInfo> OnSend;
        public event Action<Exception, string> OnError;

        public SocketUdpServer(IPAddress ipAddress, int port)
        {
            _udpClient = new UdpClient(ipAddress.AddressFamily);
            IsIPv6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            _udpClient.Client.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
            _udpClient.Client.Bind(new IPEndPoint(ipAddress, port));
        }

        public async Task StartAsync(CancellationToken cancellationToken = default)
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    var result = await _udpClient.ReceiveAsync().ConfigureAwait(false);
                    var remoteInfo = new RemoteInfo
                    {
                        EndPoint = result.RemoteEndPoint,
                        IsIPv6 = result.RemoteEndPoint.AddressFamily == AddressFamily.InterNetworkV6,
                        Data = result.Buffer
                    };

                    OnReceive?.Invoke(remoteInfo);
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                }
                catch (ObjectDisposedException ex)
                {
                    OnError?.Invoke(ex, $"ObjectDisposedException: {ex.Message}");
                    break;
                }
                catch (Exception ex)
                {
                    OnError?.Invoke(ex, $"Exception: {ex.Message}");
                }
            }
        }

        public async Task SendToAsync(IPEndPoint endPoint, byte[] data)
        {
            try
            {
                await _udpClient.SendAsync(data, data.Length, endPoint).ConfigureAwait(false);
                OnSend?.Invoke(new RemoteInfo
                {
                    EndPoint = endPoint,
                    Data = data,
                    IsIPv6 = endPoint.AddressFamily == AddressFamily.InterNetworkV6
                });
            }
            catch (SocketException ex)
            {
                OnError?.Invoke(ex, $"SocketException: {ex.Message}");
            }
            catch (ObjectDisposedException ex)
            {
                OnError?.Invoke(ex, $"ObjectDisposedException: {ex.Message}");
            }
            catch (Exception ex)
            {
                OnError?.Invoke(ex, $"Exception: {ex.Message}");
            }
        }

        public async Task SendToAllAsync(byte[] data)
        {
            var ipProperties = IPGlobalProperties.GetIPGlobalProperties();
            var endPoints = ipProperties.GetActiveUdpListeners();
            foreach (var endPoint in endPoints)
            {
                await SendToAsync(endPoint, data).ConfigureAwait(false);
            }
        }
    }
}
